/**
 * The Paper Tape Project -- Font subproject
 * PaperTapeFont C Library / Stand alone frontend
 *
 * This file contains library routines (see papertapefont.h) for
 * reading and writing "papertape font" files, as well as a
 * stand alone program.
 *
 * To compile the CLI program, define STANDALONE, e.g.
 * with
 * 
 *   gcc papertapefont.c -DSTANDALONE=1
 *
 * Initially written in a few hours (500 lines code) between
 * 02:00 and 06:00 at 04.09.2008
 *
 * (c) 2008 Sven Kppel
 *
 * This program is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software
 * Foundation; either version 3 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, see
 * http://www.gnu.org/licenses/.
 *
 **/

#include "papertapefont.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PAPERTAPE_FONT_MAX_LINE_LENGTH  100

byte_t papertape_font_line2byte(const char *line);

PAPERTAPE_FONT* papertape_font_new_from_file(FILE* fh) {
	PAPERTAPE_FONT* font;
	char line[PAPERTAPE_FONT_MAX_LINE_LENGTH];
	char ascii_id;
	byte_t* bytes;
	int bytes_n;
	int x;

	// allocate new font object
	font = malloc(sizeof(PAPERTAPE_FONT));
	if(!font || !fh)
		return NULL;

	// set up as ensurance, against segfaults or simple fonts:
	font->ascii_id = PAPERTAPE_FONT_CHARACTER_SPACING_ASCII_ID;
	font->bytes = PAPERTAPE_FONT_CHARACTER_SPACING_DEFAULT_BYTES;
	font->bytes_n = sizeof(PAPERTAPE_FONT_CHARACTER_SPACING_DEFAULT_BYTES);
	
	/*{
		byte_t[] def_spacing = PAPERTAPE_FONT_CHARACTER_SPACING_DEFAULT_BYTES;
		papertape_font_set_char(font, PAPERTAPE_FONT_CHARACTER_SPACING_ASCII_ID,
			def_spacing, sizeof(def_spacing));
		
	}*/

	// now parse file.

	// for each character package...
	while(fgets(line, PAPERTAPE_FONT_MAX_LINE_LENGTH, fh)) {
		// skip comment lines. They are only allowed at
		// index positions, not in the data byte lines.
		if(line[0] == '#')
			continue;

		ascii_id = line[0];
		bytes_n = atoi(&line[1]); // atoi will remove spaces
		bytes = calloc(sizeof(byte_t), bytes_n); // allocate place for bytes

		// parse the bytes notation
		for(x=0; x < bytes_n; x++) {
			// break at end of line
			if(!fgets(line, PAPERTAPE_FONT_MAX_LINE_LENGTH, fh))
				break;
			// add parsed line to byte array
			bytes[x] = papertape_font_line2byte(line);
		}

		papertape_font_set_char(font, ascii_id, bytes, bytes_n);
		// papertape_font_set_char copies the byte, so we free them here
		free(bytes);
	}
}

/**
 * Parse a line to a byte. This is only used internally.
 * Yeah, and it would be more performant to define this as a Macro, but that
 * can be performed using gcc -O2, too.
 **/
byte_t papertape_font_line2byte(const char *line) {
	byte_t value = 0;
	#define isPunched(pos, add) if(line[pos] != ' ') { value += add; }
	isPunched(0 ,  0);
	isPunched(1 ,  1);
	isPunched(2,   2);
	isPunched(3,   4);
	isPunched(4,   8);
	isPunched(5,  16);
	isPunched(6,  32);
	isPunched(7,  64);
	isPunched(8, 128);
	return value;
}


/**
 * This is the main function how to add or change a character in the papertape_font.
 * Internally, the papertape_font works with a ascii ordered single linked list. So
 * there's no other write access to that list than this function.
 *
 * @param font The Papertape Font object, e.g. made by papertape_font_new_...()
 * @param ascii_id The ASCII value of the character you want to add.
 * @param bytes Your byte array which is supposed to represent the character. The function copies it.
 * @param bytes_n The length of your byte array (counting from 1, of course)
 * @return 0 if successfully added, 1 otherwise. (It will never return 1 ;-) )
 **/
int papertape_font_set_char(PAPERTAPE_FONT* font, char ascii_id, const byte_t* bytes, int bytes_n) {
	PAPERTAPE_FONT_CHAR *current;
	PAPERTAPE_FONT_CHAR *next;
	
	// allocate the new character structure
	PAPERTAPE_FONT_CHAR *buf;
	buf = malloc(sizeof(PAPERTAPE_FONT_CHAR));

	// copy all values to structure
	buf->ascii_id = ascii_id;
	buf->bytes_n = bytes_n;
	buf->bytes = malloc(bytes_n);
	memcpy(buf->bytes, bytes, bytes_n);
	buf->next = NULL;

	// Magic ascii id (typically a non printable character) for character spacing
	if(ascii_id == PAPERTAPE_FONT_CHARACTER_SPACING_ASCII_ID) {
		font->bytes = buf->bytes;
		font->bytes_n = buf->bytes_n;
		free(buf); // don't need it.
		return 0;
	}

	// Start going throught the list
	current = font; // Start with the font object as first list member.
	while(current->next != NULL) {
		if(current->next->ascii_id == ascii_id) {
			// the next one has the same ascii_id as we have (perhaps
			// we are 'c' and the next is 'c', too). So replace the next
			// one.
			next = current->next;  // copy the pointer somewhere else
			current->next = buf;   // replace the next pointer with our char
			free(next);            // delete the old next pointer.
			return 0;
		} else if(current->next->ascii_id > ascii_id) {
			// the next one has a bigger ascii_id as we have (perhapse
			// we are 'c' while the next is 'e'). So pack our character
			// right here, between current and next.
			next = current->next;  // copy the pointer somewhere else
			current->next = buf;   // replace the next pointer with our char
			buf->next = next;      // pack the next pointer after our pointer
			return 0;
		} else {
			// the next one has a lower ascii_id as we have, so we have
			// to go on throught the list
			current = current->next;
		}
	}

	// we have reached the end of the list and haven't found any ascii_id biggger
	// or equal to ours. So we simply add our character to the list.
	current->next = buf;
	return 0;
}

/**
 * This is the main function from the papertape_font package. Using this function, you
 * can render any ASCII string to a papertape label, using the current font.
 *
 * The return byte array will be malloced, and you'll have after calling the only
 * link to that array. So feel free to free() it, if you don't need it any more.
 *
 * If there's a non printable character in your string, this function will print out an
 * error message at stderr and skip that character. Use papertape_font_string_is_printable
 * if you want to check wheter your string contains non printable characters.
 *
 * Internally speaken, this is the only read access function to the internal single
 * linked font character list.
 *
 **/
byte_t* papertape_font_get_label(PAPERTAPE_FONT* font, const char* string) {
	PAPERTAPE_FONT_CHAR* current;
	//int string_len = strlen(string);
	int x; // to count throught the string
	int size = 0; // the size of the output byte array
	byte_t* output;
	int pos; // current position in output array

	// first step: Count the neccessary size for the output byte array
	for(x=0; string[x] != '\0'; x++) {
		// skip the font spacing at one character in the string
		// (because it's only between the letters)
		if(x != 0)
			size += font->bytes_n;

		// go throught the list and try to find the current character.
		current = font;
		while(1) {
			if(current->ascii_id == string[x]) {
				size += current->bytes_n;
				break;
			} else if(current->next != NULL) {
				current = current->next;
				continue;
			} else {
				// We have reached the end of the character
				// list, but the current character is not available!
				fprintf(stderr, "papertape_font_get_label: Character \"%c\" in \"%s\" not printable\n",
					string[x], string);
				break;
			}
		}
	}

	// now malloc the output byte array and generate it
	output = malloc(size);
	for(x=0,pos=0; string[x] != '\0'; x++) {
		// go throught the list, another time
		current = font;
		while(1) {
			if(current->ascii_id == string[x]) {
				memcpy(output+pos, current->bytes, current->bytes_n);
				pos += current->bytes_n;
				break;
			} else if(current->next != NULL) {
				current = current->next;
				continue;
			} // else: Character not found. Ignore it,
			//   because error already printed at first run.
		}
		// add the character spacer after each character
		if(x+1 < strlen(string)) { // but not at the last character
			memcpy(output+pos, font->bytes, font->bytes_n);
			pos += font->bytes_n;
		}
	}

	// we are done.
	return output;
}

/**
 * Checks if there's any non printable character in your string. Papertape fonts don't have
 * to have bit masks for every character, so use this function to check your string.
 * papertape_font_get_label will output errors to stderr if it finds unprintable
 * characters.
 *
 * @returns 1 if all characters are printable, 0 otherwise.
 **/
int papertape_font_string_is_printable(PAPERTAPE_FONT *font, const char* string) {
	PAPERTAPE_FONT_CHAR* current;
	int x;

	for(x=0; string[x] != '\0'; x++) {
		current = font;
		while(1) {
			if(current->ascii_id == string[x])
				// this character is clean.
				break;
			else if(current->next != NULL)
				// have not found character yet.
				current = current->next;
			else
				// at the end of list -- character not found!
				return 0;
		}
	}

	return 1; // No non printable char found.
}

/**
 * Write the current PAPERTAPE_FONT (back) out to a file.
 * FILE mus be a writeable handle.
 **/
int papertape_font_write_to_file(PAPERTAPE_FONT* font, FILE *fh) {
	PAPERTAPE_FONT_CHAR* current;
	int x;

	// Write first lines.
	fputs(PAPERTAPE_FONT_WRITE_FIRST_LINES, fh);
	
	// now go throught list and write all characters out.
	current = font;
	while(1) {
		// write index line
		fprintf(fh, "%c %i\n", current->ascii_id, current->bytes_n);
		// write data binary things.
		for(x=0; x<current->bytes_n; x++) {
			#define paintPunched(check) fputc((current->bytes[x] & check) ?\
				fputc(PAPERTAPE_FONT_WRITE_LOGICAL_1_CHARACTER, fh) : fputc(' ', fh), fh)
			paintPunched(  0);
			paintPunched(  1);
			paintPunched(  2);
			paintPunched(  4);
			paintPunched(  8);
			paintPunched( 16);
			paintPunched( 32);
			paintPunched( 64);
			paintPunched(128);
			fputc('\n', fh);
		}

		// and go on. Or break, if at end.
		if(current->next != NULL)
			current = current->next;
		else
			break;
	}
}

/* If we want it to be a standalone program */
#ifdef STANDALONE
int file_get_contents(FILE *stream, byte_t **content) {
	byte_t buf[4096];
	size_t bytes; // gerade eben eingelesene bytes
	byte_t *str = NULL;
	size_t total_bytes = 0;     // alle bis jetzt eingelesenen bytes
	size_t total_allocated = 0;
	byte_t *tmp; // fuers realloc

	while(!feof(stream)) {
		bytes = fread(buf, 1, sizeof(buf), stream);

		while( (total_bytes + bytes) > total_allocated) {
			if(str)
				total_allocated *= 2;
			else
				total_allocated = bytes > sizeof(buf) ? bytes : sizeof(buf);

			tmp = realloc(str, total_allocated);

			if(tmp == NULL) {
				fprintf(stderr, "*** file_get_contents ERROR*** Could not reallocate\n");
				//return length; // Fehler - das eingelesene zumindest zurueckgeben.
				// nee, gar nichts zurckgeben. Das geht so nicht.
				return 0;
			}

			str = tmp;
		} // while innen

		memcpy(str + total_bytes, buf, bytes);
		total_bytes += bytes;
	} // while

	if(total_allocated == 0)
		str = malloc(1); // something empty. Just to be not NULL...
	//str[total_bytes] = '\0'; // wenns ein string wre.

	*content = str;
	return total_bytes;
}

int main(int argc, char **argv) {
	enum {
		RENDER_TEXT,
		WRITE_CHAR,
		EXPORT,
		IMPORT,
		NOTHING
	} action;
	action = NOTHING;
	char *schematics = PAPERTAPE_FONT_SCHEMATICS_DEFAULT;
	char *render_text;
	char write_char;
	char *font_name;
	FILE *font_file;
	PAPERTAPE_FONT *font;
	int c;
	
	// poor man's getopt, for windows portability
	for(c=1; c<argc; c++) {
		if(strlen(argv[c]) == 2 && *argv[c] == '-') {
			argv[c]++; // shift "-" flag thingy.
			switch(*argv[c]) {
				case 'h':
					fprintf(stderr, "Usage: %s [OPTIONS...] [A PAPERTAPE FONT FILE]\n"
						"The Papertape font utilitie program.\n\n"
						"   optional parameters:\n"
						"   -h              display this help message\n"
						"   -l [TEXT]       Render this text to STDOUT\n"
						"   -w [ASCII ID]   Set the character with that ASCII ID, read from STDIN\n"
						"   -o              Export the font file, according to schematics, to STDOUT\n"
						"   -i              Import to font file (overwrite it), according to schematics, from STDIN\n\n"
						"   -s [Sch string] Define own schematics for -i/-o. Default is:\n"
						"                   %s\n",
						argv[0], PAPERTAPE_FONT_SCHEMATICS_DEFAULT);
					return 0;

				case 'l':
					action = RENDER_TEXT;
					render_text = argv[++c];
					break;
				
				case 'w':
					action = WRITE_CHAR;
					write_char = *argv[++c];
					break;
					
				case 'o': action = EXPORT; break;
				case 'i': action = IMPORT;  break;
				
				case 's':
					schematics = argv[++c];
			} // switch
		} // if flag
		else {
			// not flag but... Papertape Font file!
			font_name = argv[c];
		}
	} // for each argv

	// Open Font File
	if(font_name) {
		char *mode;
		switch(action) {
			case RENDER_TEXT:
			case EXPORT:       mode = "r";  break;
			case WRITE_CHAR:   mode = "rw"; break;
			case IMPORT:       mode = "w";  break;
			default:
				fprintf(stderr, "Error: No action specified. Call %s -h\n", argv[0]);
				return 1;
		}
	
		font_file = fopen(font_name, mode);
		if(!font_file) {
			perror("Opening font file");
			return 2;
		}
	} else {
		fprintf(stderr, "Error: No font file specified. Call %s -h\n", argv[0]);
	}
	
	// Stupid: papertapefont.c:431: Warnung: Zuweisung erzeugt Zeiger von Ganzzahl ohne Typkonvertierung
	// Adding (PAPERTAPE_FONT*) did the job.
	font = (PAPERTAPE_FONT*) papertape_new_from_file(font_file);
	if(font == NULL) {
		fprintf(stderr, "Error creating PAPERTAPE_FONT object\n");
		return 3;
	}

	// Perform action
	if(action == RENDER_TEXT && render_text != NULL) {
		papertape_font_get_label(font, render_text);
	} else if(action == WRITE_CHAR) {
		byte_t *contents;
		int bytes = file_get_contents(stdin, &contents);
		printf("Read in %i bytes for letter %c (seen as byte: 0x%x)\n", bytes, write_char, write_char);
		papertape_font_set_char(font, write_char, contents, bytes);
	} else if(action == EXPORT) {
		fprintf(stderr, "Printing file %s with schematics %s\n", font_name, schematics);
		puts(papertape_font_get_label(font, schematics));
	} else if(action == IMPORT) {
		byte_t *contents;
		int bytes, x;
		
		fprintf(stderr, "Reading from stdin, expecting schematics %s\n", schematics);
		bytes = file_get_contents(stdin, &contents);
		
		fprintf(stderr, "To be implemented...\n");
		
		// Parse schematics
		/*for(x=0; x<strlen(schematics); x++) {
			papertape_font_set_char(font, schematics[x], 
		}*/
		
	} else {
		fprintf(stderr, "No correct call! See %s -h\n", argv[0]);
	}
	return 0;
}
#endif /* STANDALONE */
